%option noyywrap
%option nostdinit
%option yylineno
%option case-insensitive
%option full

%{
%}

LETTER [a-zA-Z]
DIGIT [0-9]

%%

"("|")"|","|[ \t\n]				/* eat whitespace */;
";".*							{                              return SEMICOLON;}

"ADD"							{yylval.number = ADD;          return OP;}
"SUB"							{yylval.number = SUB;          return OP;}
"MUL"|"MULT"					{yylval.number = MUL;          return OP;}
"DIV"							{yylval.number = DIV;          return OP;}
"MOD"							{yylval.number = MOD;          return OP;}

"LT"							{yylval.number = LT;           return OP;}
"LE"							{yylval.number = LE;           return OP;}
"EQ"							{yylval.number = EQ;           return OP;}
"GE"							{yylval.number = GE;           return OP;}
"GT"							{yylval.number = GT;           return OP;}
"NEQ"							{yylval.number = NEQ;          return OP;}

"AND"							{yylval.number = AND;          return OP;}
"OR"							{yylval.number = OR;           return OP;}
"NOT"							{yylval.number = NOT;          return OP;}

"JMP"							{yylval.number = JMP;          return OP;}
"JFALSE"						{yylval.number = JFALSE;       return OP;}
"CALL"							{yylval.number = CALL;         return OP;}
"RET"							{yylval.number = RET;          return OP;}

"PUSH"							{yylval.number = PUSH;         return OP;}
"POP"							{yylval.number = POP;          return OP;}
"FP"							{                              return FP;}
"R"								{                              return R;}
"SP"							{                              return SP;}
"<"								{                              return REF;}
">"								{                              return REFCLOSE;}
"+"								{                              return ADD;}
"-"								{                              return SUB;}
"*"								{                              return MUL;}
"/"								{                              return DIV;}
"%"								{                              return MOD;}


"LIT"							{yylval.number = LIT;          return OP;}
"LOAD"							{yylval.number = LOAD;         return OP;}
"STO"|"STOR"|"STORE"			{yylval.number = STORE;        return OP;}

{DIGIT}+						{yylval.number = atoi(yytext); return NUM;}
({DIGIT}|"_"|{LETTER})+			{yylval.str = yytext;          return LABEL;}
":"								{                              return LABELDEF;}
.								{cerr << "__LEXICAL ERROR (line " << yylineno << "): unexpected character \""<<yytext[0]<<"\"\n";}

%%