
#include<iostream>
#include<fstream>
#include<vector>
#include<map>
using namespace std;

int stack_max = 1024*1024;
vector<int> stack=vector<int>(0);
int stack_pointer=-1;
int frame_pointer=3;
int instruction_counter=1;
int index_register;
map<string,int> jump_labels;
int jump_label = 0;
vector<int> jump_line;
int codeline=1;

int pop() {
	return stack[stack_pointer--];
}
void push(int val) {
	if(stack_pointer >= stack_max) {
		cerr << "Stack overflow\n";
		exit(1);
	}
	stack[++stack_pointer] = val;
}
void ShowStack(ostream &os) {
	int frame_sp = stack_pointer;
	int frame_fp = frame_pointer;
	while(frame_fp > 0) {
		if(frame_sp > frame_fp) {
			os << stack[frame_sp--];
			while(frame_sp > frame_fp)
				os << ":" << stack[frame_sp--];
		}
		os << "[" << stack[frame_sp] << ":" << stack[frame_sp-1]
		   << ":" << stack[frame_sp-2] << "]  ";
		frame_sp -= 3;
		frame_fp = stack[frame_fp];
	}

	if(frame_sp >= 0) {
		os << stack[frame_sp--];
		while(frame_sp >= 0)
			os << ":" << stack[frame_sp--];
	}
}

#include "parser.tab.c"

int main(int argc, char** argv) {
	bool showops = false;
	bool showstack = false;
	bool creep = false;

	int i = 1;
	string argvi;
	for(; i < argc; i++) {
		argvi = argv[i];
		if(argvi == "-s" || argvi == "--show-stack")
			showstack = true;
		else if(argvi == "-o" || argvi == "--show-ops")
			showops = true;
		else if(argvi == "-c" || argvi == "--creep")
			creep = true;
		else if(argvi == "--stack-size") {
			if(i < argc-1)
				stack_max = atoi(argv[i+1]);
			i++;
		}
		else if(argvi == "-h" || argvi == "--help" || argvi == "-help" || argvi == "--h") {
		cout << "\n +----------------------------------------------------------------------------------+\n"
		     << " | Virtual Machine for the example intermediate-code \"ZPP\" as                       |\n"
		     << " | taught in the \"compiler-construction\" lectures at the University of Technology   |\n"
		     << " | Aachen (Germany)                written by Viktor \"AlgorithMan\" Engelmann 2009   |\n"
		     << " +----------------------------------------------------------------------------------+\n"
		     << " | this program is provided under the terms of the GNU general public license (GPL) |\n"
		     << " | Version 3. See gpl3.txt for details.                                             |\n"
		     << " +----------------------------------------------------------------------------------+\n\n";


			cout << " usage: " << argv[0] << " [options] <bps-file> [parameters for bps program]\n";
			cout << " where [options] can be\n";
			cout << "       -s --show-stack    to show the stack after every operation\n";
			cout << "       -o --show-ops      to show every operation before executing it\n";
			cout << "       -c --creep         to pause after every operation (note that\n";
			cout << "                          this option is disabled, if neither stack\n";
			cout << "                          nor ops are shown)\n";
			cout << "       --stack-size <n>   to change the size of the stack to <n>\n\n";
			cout << " the [parameters for bps program] must be integers and are put on the\n";
			cout << " stack of the main frame (the last parameter is pushed first, then the)\n";
			cout << " but-last and so on)\n";
			cout << " EXAMPLE: " << argv[0] << " -s -o -c fac_recursive.zpp 4 3\n";
			cout << " would start with the main frame 0:0:0:4:3\n\n";
			return 1;
		} else
			break;
	}

	if(!showstack && !showops)
		creep = false;
	if(i >= argc) {
		cerr << argv[0] << ": No ZPP file specified.\n";
		return 1;
	}


	cout << "LOADING PROGRAM...";
	jump_labels["0"] = 0;
	jump_line.push_back(0);
	jump_label = 1;
	yyin = fopen(argv[i], "r");
	yyparse();
	if(yynerrs != 0) {
		cerr << " " << yynerrs << " syntactic errors occured\n";
		return 1;
	} else
		cout << " DONE\n";



	cout << "PREPARING VIRTUAL MACHINE...";
	stack = vector<int>(stack_max,0);
	for(int j = argc - 1; j > i; --j)
		push(atoi(argv[j]));
	push(0);
	push(0);
	push(0);
	frame_pointer = stack_pointer;
	cout << "\nINITIAL STACK: ";
	ShowStack(cout);



	cout << "\nSTARTING EXECUTION\n";
	int lval;
	while(instruction_counter > 0) {

		if(showstack) {
			cout << "  ";
			if(!showops)
				cout << "{IC="<<instruction_counter << ", FP=" << frame_pointer
				     << ", SP="<<stack_pointer<<", R="<<index_register<<"}\t";
			ShowStack(cout);
			if(!creep || showops) cout << "\n";
		}
		if(showops) {
			cout << "{IC="<<instruction_counter << ", FP=" << frame_pointer
			     << ", SP="<<stack_pointer<<", R="<<index_register<<"}\t"
			     << names[program[instruction_counter]->opcode-LABEL];
			if(program[instruction_counter]->param1 != NULL)
				cout << "\t" << program[instruction_counter]->param1->tostring();
			if(program[instruction_counter]->param2 != NULL)
				cout << "\t" << program[instruction_counter]->param2->tostring();
			if(program[instruction_counter]->param3 != NULL)
				cout << "\t" << program[instruction_counter]->param3->tostring();
			if(!creep) cout << "\n";
		}
		if(creep)
			cin.ignore(1);

		switch(program[instruction_counter]->opcode) {
			case ADD:
				if(program[instruction_counter]->param1 == NULL)
					push(pop() + pop());
				else // ADD with params
					switch(program[instruction_counter]->param1->id) {
						case REF:
							stack[program[instruction_counter]->param1->param1->val()] += program[instruction_counter]->param2->val();
							break;
						case FP:
							frame_pointer += program[instruction_counter]->param2->val();
							break;
						case SP:
							stack_pointer += program[instruction_counter]->param2->val();
							break;
						case R:
							index_register += program[instruction_counter]->param2->val();
							break;
					}
				break;
			case SUB: 
				if(program[instruction_counter]->param1 == NULL) {
					lval = pop();
					push(pop() - lval);
				} else // SUB with params
					switch(program[instruction_counter]->param1->id) {
						case REF:
							stack[program[instruction_counter]->param1->param1->val()] -= program[instruction_counter]->param2->val();
							break;
						case FP:
							frame_pointer -= program[instruction_counter]->param2->val();
							break;
						case SP:
							stack_pointer -= program[instruction_counter]->param2->val();
							break;
						case R:
							index_register -= program[instruction_counter]->param2->val();
							break;
					}
				break;
			case MUL:
				push(pop() * pop());
				break;
			case DIV:
				lval = pop();
				push(pop() / lval);
				break;
			case MOD:
				lval = pop();
				push(pop() % lval);
				break;
			case LT:
				lval = pop();
				push((pop() < lval)?1:0);
				break;
			case LE:
				lval = pop();
				push((pop() <= lval)?1:0);
				break;
			case EQ:
				push((pop() == pop())?1:0);
				break;
			case GT:
				lval = pop();
				push((pop() > lval)?1:0);
				break;
			case GE:
				lval = pop();
				push((pop() >= lval)?1:0);
				break;
			case NEQ:
				push((pop() != pop())?1:0);
				break;

			case NOT:
				push((pop() == 0)?1:0);
				break;
			case AND:
				push((pop() != 0 && pop() != 0)?1:0);
				break;
			case OR:
				push((pop() != 0 || pop() != 0)?1:0);
				break;

			case JMP:
				instruction_counter = program[instruction_counter]->param1->val()-1;
				break;
			case JFALSE:
				if(pop() == 0)
					instruction_counter = program[instruction_counter]->param1->val()-1;
				break;

			case CALL:
				push(instruction_counter+1); // return address
				instruction_counter = program[instruction_counter]->param1->val()-1;
				break;
			case RET:
				lval = stack_pointer;
				stack_pointer -= program[instruction_counter]->param1->val()+1;
				instruction_counter = stack[lval]-1;
				break;

			case PUSH: case LIT:
				push(program[instruction_counter]->param1->val());
				break;
			case LOAD:
				switch(program[instruction_counter]->param1->id) {
					case REF:
						stack[program[instruction_counter]->param1->param1->val()] = program[instruction_counter]->param2->val();
						break;
					case FP:
						frame_pointer = program[instruction_counter]->param2->val();
						break;
					case SP:
						stack_pointer = program[instruction_counter]->param2->val();
						break;
					case R:
						index_register = program[instruction_counter]->param2->val();
						break;
				}
				break;
			case POP:
				switch(program[instruction_counter]->param1->id) {
					case REF:
						stack[program[instruction_counter]->param1->param1->val()] = pop();
						break;
					case FP:
						frame_pointer = pop();
						break;
					case SP:
						stack_pointer = pop();
						break;
					case R:
						index_register = pop();
						break;
				}
				break;
		}
		instruction_counter++;
	}



	cout << "\n F[" << argv[i] << "](";
	if(argc > i+1)
		cout << argv[i+1];
	for(int j = i+2; j < argc; ++j)
		cout << "," << argv[j];
	cout << ") = (";
	for(int k = stack_pointer-3; k > 0; --k)
		cout << stack[k] << ",";
	if(stack_pointer >= 3)
		cout << stack[0];
	cout << ")\n\n";


	for(int i = 0; i < program.size(); i++)
		delete(program[i]);
	return 0;
}

