%{
#define YYERROR_VERBOSE

#include<iostream>
#include<vector>
#include<map>
using namespace std;
void yyerror(char* s);
int yylex();


string itoa(int i) {
	if(i < 0)
		return "-" + itoa(-i);
	if(i == 0)
		return "0";
	string result;
	while(i > 0) {
		result = string(1, '0'+i%10) + result;
		i /= 10;
	}
	return result;
}

string names[] = {
       "LABEL", "SEMICOLON", "OP", "FP", "R", "SP", "REF", "REFCLOSE", "NUM",
       "LABELDEF",
       "ADD", "SUB", "MUL", "DIV", "MOD",
       "LT", "LE", "EQ", "GE", "GT", "NEQ",
       "AND", "OR", "NOT",
       "JMP", "JFALSE", "CALL", "RET",
       "PUSH", "POP",
       "LOAD", "LIT", "STORE"};
class exp {
	public:
		int id;
		int num;
		exp* param1;
		exp* param2;

		exp(int i, int n, exp* p1, exp* p2) {
			id = i;
			num = n;
			param1 = p1;
			param2 = p2;
		}
		~exp() {
			if(param1 != NULL)
				delete param1;
			if(param2 != NULL)
				delete param2;
		}
		int val() {
			switch(id) {
				case NUM:   return num;
				case ADD:   return param1->val() + param2->val();
				case SUB:   return param1->val() - param2->val();
				case MUL:   return param1->val() * param2->val();
				case DIV:   return param1->val() / param2->val();
				case MOD:   return param1->val() % param2->val();
				case REF:   return stack[param1->val()];
				case SP:    return stack_pointer;
				case FP:    return frame_pointer;
				case R:     return index_register;
				case LABEL: return jump_line[num];
			}
		}
		string tostring() {
			switch(id) {
				case NUM:   return itoa(num);
				case ADD:   return "("+param1->tostring()+"+"+param2->tostring()+")";
				case SUB:   return "("+param1->tostring()+"-"+param2->tostring()+")";
				case MUL:   return "("+param1->tostring()+"*"+param2->tostring()+")";
				case DIV:   return "("+param1->tostring()+"/"+param2->tostring()+")";
				case MOD:   return "("+param1->tostring()+"%"+param2->tostring()+")";
				case REF:   return "<"+param1->tostring()+">";
				case SP:    return "SP";
				case FP:    return "FP";
				case R:     return "R";
				case LABEL: return itoa(jump_line[num]);
			}
		}
		static exp* number(int n)            {return new exp(NUM,n,NULL,NULL);}
		static exp* plus(exp* p1, exp* p2)   {return new exp(ADD,0,p1,p2);}
		static exp* minus(exp* p1, exp* p2)  {return new exp(SUB,0,p1,p2);}
		static exp* times(exp* p1, exp* p2)  {return new exp(MUL,0,p1,p2);}
		static exp* divide(exp* p1, exp* p2) {return new exp(DIV,0,p1,p2);}
		static exp* mod(exp* p1, exp* p2)    {return new exp(MOD,0,p1,p2);}
		static exp* ref(exp* p1)             {return new exp(REF,0,p1,NULL);}
		static exp* sp()                     {return new exp(SP,0,NULL,NULL);}
		static exp* fp()                     {return new exp(FP,0,NULL,NULL);}
		static exp* r()                      {return new exp(R,0,NULL,NULL);}
		static exp* label(int idx)           {return new exp(LABEL,idx,NULL,NULL);}

};
class op {
	public:
		int opcode;
		exp* param1;
		exp* param2;
		exp* param3;
		op(int o, exp* p1, exp* p2, exp* p3) {
			opcode = o;
			param1 = p1;
			param2 = p2;
			param3 = p3;
		}
		~op() {
			if(param1 != NULL)
				delete(param1);
			if(param2 != NULL)
				delete(param2);
			if(param3 != NULL)
				delete(param3);
		}
};


vector<op*> program(1,(op*)(NULL));

%} // Yacc Deklarationen


%union {exp* expression; int number; char* str;}
%start S
%token <str> LABEL
%token <number> SEMICOLON OP FP R SP REF REFCLOSE NUM
       LABELDEF
       ADD SUB MUL DIV MOD
       LT LE EQ GE GT NEQ
       AND OR NOT
       JMP JFALSE CALL RET
       PUSH POP 
       LOAD LIT STORE
%left ADD SUB
%left MUL DIV MOD

%% // Produktionen

S   : | SEMICOLON S | OPERATION S
    ;
OPERATION : OP EXP EXP EXP SEMICOLON {
                program.push_back(new op($<number>1, $<expression>2, $<expression>3, $<expression>4));
                codeline++;
             }
          | OP EXP EXP SEMICOLON {
                program.push_back(new op($<number>1, $<expression>2, $<expression>3, NULL));
                codeline++;
             }
          | OP EXP SEMICOLON {
                program.push_back(new op($<number>1, $<expression>2, NULL, NULL));
                codeline++;
             }
          | OP SEMICOLON {
                program.push_back(new op($<number>1, NULL, NULL, NULL));
                codeline++;
             }
          | LABEL {
                if(jump_labels.find($1) == jump_labels.end()) {
                   jump_labels[$1] = jump_label++;
                   jump_line.push_back(codeline);
                } else {
                   if(jump_line[jump_labels[$1]] == -2)
                      jump_line[jump_labels[$1]] = codeline;
                   else
                      cerr << "jump label \"" << $1 << "\" already defined\n";
                }
             } LABELDEF
          | NUM {
                string temp = itoa($1);
                if(jump_labels.find(temp) == jump_labels.end()) {
                   jump_labels[temp] = jump_label++;
                   jump_line.push_back(codeline);
                } else {
                   if(jump_line[jump_labels[temp]] == -2)
                      jump_line[jump_labels[temp]] = codeline;
                   else
                      cerr << "jump label \"" << temp << "\" already defined\n";
                }
             } LABELDEF
          ;
EXP : NUM              {$<expression>$ = exp::number($1);}
    | EXP ADD EXP      {$<expression>$ = exp::plus($<expression>1,$<expression>3);}
    | EXP SUB EXP      {$<expression>$ = exp::minus($<expression>1,$<expression>3);}
    | EXP MUL EXP      {$<expression>$ = exp::times($<expression>1,$<expression>3);}
    | EXP DIV EXP      {$<expression>$ = exp::divide($<expression>1,$<expression>3);}
    | EXP MOD EXP      {$<expression>$ = exp::mod($<expression>1,$<expression>3);}
    | REF EXP REFCLOSE {$<expression>$ = exp::ref($<expression>2);}
    | FP               {$<expression>$ = exp::fp();}
    | SP               {$<expression>$ = exp::sp();}
    | R                {$<expression>$ = exp::r();}
    | LABEL            {
          if(jump_labels.find($1) == jump_labels.end()) {
             jump_labels[$1] = jump_label++;
             jump_line.push_back(-2);
          }
          $<expression>$ = exp::label(jump_labels[$1]);
       }
    ;

%%

#include "lex.yy.c"

void yyerror(char* s) {
	cerr << "SYNTACTIC ERROR (Line " << yylineno << "): " << s << endl;
}
/*  */
